/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : SCPCrossover1PX.cc                                       */
/*                                                                       */
/*       Date : March 5, 1998                                            */
/* Updated at : March 5, 1998                                            */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/*************************************************************************/

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>

#include <Chromosomes.h>
#include <ChromosomesBag.h>
#include <WeightedRandomPickerSet.h>
#include <tSet.h>

#include "SCPCrossover1PX.h"
#include "SCPChromosomes.h"
#include "SetCoverProblem.h"
#include "SetCoverSolution.h"

/*************************************************************************/
/*************************************************************************/

SCPCrossover1PX::SCPCrossover1PX() { debugOn(); }

/*************************************************************************/
/*************************************************************************/

SCPCrossover1PX::~SCPCrossover1PX()
{  cout << " de allocate SCPCrossover1PX " << endl; }

/*************************************************************************/
/*************************************************************************/

ChromosomesBag SCPCrossover1PX::apply(Chromosomes* par1, Chromosomes * par2)
{   
    if(debug) cout << " I am in SCPCrossover1PX::apply() " << endl;
    SCPChromosomes *parent1 = (SCPChromosomes *) par1;
    SCPChromosomes *parent2 = (SCPChromosomes *) par2;
    SetCoverProblem *problem = parent1->getProblem();
    int numOfRows = problem->getNumOfRows();
    int numOfCols = problem->getNumOfCols();
    int i, j, left, right, pxindex ;

    /*
    Column cpr1(numOfCols);
    Column cpr2(numOfCols);
    Column coff1(numOfCols);
    Column coff2(numOfCols);
    */

    if(debug) cout << " 1. I am in SCPCrossover1PX::apply() " << endl;

    SCPChromosomes *offspring1 = new SCPChromosomes(parent1->getProblem());
    SCPChromosomes *offspring2 = new SCPChromosomes(parent1->getProblem());

    if(debug) cout << " 2. I am in SCPCrossover1PX::apply() " << endl;

    // Build binary representation
    tSet<int> a1 = parent1->getColumns();

    int a1Size = a1.getSize();
    int *a1Indices = new int[a1Size];
    for(i=0;i<a1Size;i++) if( a1.getElementAtSafe(i,j)) a1Indices[i]=j;

    a1 = parent2->getColumns();
    int a2Size = a1.getSize();
    int *a2Indices = new int[a2Size];
    for(i=0;i<a2Size;i++) if( a1.getElementAtSafe(i,j)) a2Indices[i]=j;

    int dummy;

    // sort non-decreasing
    for(i=0;i<a1Size-1;i++)
      for(j=i+1;j<a1Size;j++)
	if(a1Indices[i]>a1Indices[j])
	{
	   dummy = a1Indices[j];
	   a1Indices[j] = a1Indices[i];
	   a1Indices[i] = dummy;
	}

    // sort non-decreasing
    for(i=0;i<a2Size-1;i++)
      for(j=i+1;j<a2Size;j++)
	if(a2Indices[i]>a2Indices[j])
	{
	   dummy = a2Indices[j];
	   a2Indices[j] = a2Indices[i];
	   a2Indices[i] = dummy;
	}

    i = 0;
    while( i<numOfRows && i<a1Size && 
	   i<a2Size && a1Indices[i] == a2Indices[i] ) i++; 
    if( i<a1Size && i<a2Size) left = minimum(a1Indices[i],a2Indices[i]);
    else if( i==a1Size && a1Size==a2Size) 
	 left = minimum(a1Indices[i-1],a2Indices[i-1]);
    else if(i>=a1Size) left = a2Indices[i];
    else if(i>=a2Size) left = a1Indices[i];

    i = a1Size-1;
    j = a2Size-1;
    while( i>=0 && j>=0 && a1Indices[i] == a2Indices[j] ) { i--;j--;}

    if( i>=0 && j>=0 ) right = minimum(a1Indices[i],a2Indices[j]);
    else if( i<0 && j<0 ) right = minimum(a1Indices[0],a2Indices[0]);
    else if( i<0 ) right = a2Indices[j];
    else if( j<0 ) right = a1Indices[i];

    if(left>right) { pxindex=right; right=left; left=pxindex;} 
    pxindex = rand()%(right-left+1)+left;


    // put all elements less than pxindex
    i=0;
    while( i<a1Size && a1Indices[i]<pxindex) 
    {  offspring1->addColumn(a1Indices[i]); i++; }

    j=0;
    while( j<a2Size && a2Indices[j]<pxindex) 
    {  offspring2->addColumn(a2Indices[j]); j++; }

    // exchange
    while( i<a1Size ) 
    {  offspring2->addColumn(a1Indices[i]); i++; }

    while( j<a2Size ) 
    {  offspring1->addColumn(a2Indices[j]); j++; }

    /*
    // put all elements less than left
    i=0;
    while( i<a1Size && a1Indices[i]<left) 
    {  offspring1->addColumn(a1Indices[i]); i++; }

    j=0;
    while( j<a2Size && a2Indices[j]<left) 
    {  offspring2->addColumn(a2Indices[j]); j++; }

    while( i<a1Size && a1Indices[i]<right) 
    {  offspring2->addColumn(a1Indices[i]); i++; }

    while( j<a2Size && a2Indices[j]<right) 
    {  offspring1->addColumn(a2Indices[j]); j++; }

    // put all elements greater than right
    while( i<a1Size ) {  offspring1->addColumn(a1Indices[i]); i++; }

    while( j<a2Size ) {  offspring2->addColumn(a2Indices[j]); j++; }
    */

    repair(offspring1);
    repair(offspring2);

    /*if(debug) 
    { 
       cout << " parent1   : " << *parent1 << endl;
       cout << " parent2   : " << *parent2 << endl;
       cout << " offspring1 : " << *offspring1 << endl;
       cout << " offspring2 : " << *offspring2 << endl;

    }// end of if
    */

    offspring1->reduce();
    offspring2->reduce();

    /*
    if(debug) 
    {
      ofstream ou1("/tmp/1px.out",ios::app);
      ou1  << " parent1   : " << *parent1 << endl;
      ou1  << " parent2   : " << *parent2 << endl;
      ou1  << " offspring1 : " << *offspring1 << endl;
      ou1  << " offspring2 : " << *offspring2 << endl;
      ou1.close();

    }// end of if
    */

    ChromosomesBag ret(2);
    ret.addChromosomes(offspring1);
    ret.addChromosomes(offspring2);

    if(debug) cout << " 3. I am out of SCPCrossover1PX::apply() " << endl;

    delete[] a1Indices;
    delete[] a2Indices;

    return ret;

}// end of apply()

/*************************************************************************/
/*************************************************************************/

ChromosomesBag SCPCrossover1PX::applyOLD(Chromosomes* par1, Chromosomes * par2)
{   
    if(debug) cout << " I am in SCPCrossover1PX::apply() " << endl;
    SCPChromosomes *parent1 = (SCPChromosomes *) par1;
    SCPChromosomes *parent2 = (SCPChromosomes *) par2;
    SetCoverProblem *problem = parent1->getProblem();
    int numOfRows = problem->getNumOfRows();
    int numOfCols = problem->getNumOfCols();
    int i, j, left, right, pxindex ;

    Column cpr1(numOfCols);
    Column cpr2(numOfCols);
    Column coff1(numOfCols);
    Column coff2(numOfCols);

    if(debug) cout << " 1. I am in SCPCrossover1PX::apply() " << endl;

    SCPChromosomes *offspring1 = new SCPChromosomes(parent1->getProblem());
    SCPChromosomes *offspring2 = new SCPChromosomes(parent1->getProblem());

    if(debug) cout << " 2. I am in SCPCrossover1PX::apply() " << endl;

    // Build binary representation
    tSet<int> a1 = parent1->getColumns();
    for(i=0;i<a1.getSize();i++) 
	if( a1.getElementAtSafe(i,j)) cpr1.set(j);

    a1 = parent2->getColumns();
    for(i=0;i<a1.getSize();i++) 
	if( a1.getElementAtSafe(i,j)) cpr2.set(j);

    left  = findFirstLeftDifferent(&cpr1,&cpr2,numOfCols);
    right = findFirstRightDifferent(&cpr1,&cpr2,numOfCols);

    if(left>right) { pxindex=right; right=left; left=pxindex;} 
    pxindex = rand()%(right-left+1)+left;

    for(i=0;i<=pxindex;i++) 
    {
       if( cpr1.test(i) ) coff1.set(i);
       if( cpr2.test(i) ) coff2.set(i);

    }// end of for

    for(;i<numOfCols;i++) 
    {
       if( cpr1.test(i) ) coff2.set(i);
       if( cpr2.test(i) ) coff1.set(i);

    }// end of for

    for(i=0;i<numOfCols;i++) if( coff1.test(i) ) offspring1->addColumn(i);

    for(i=0;i<numOfCols;i++) if( coff2.test(i) ) offspring2->addColumn(i);

    repair(offspring1);
    repair(offspring2);

    /*if(debug) 
    { 
       cout << " parent1   : " << *parent1 << endl;
       cout << " parent2   : " << *parent2 << endl;
       cout << " offspring1 : " << *offspring1 << endl;
       cout << " offspring2 : " << *offspring2 << endl;

    }// end of if
    */

    offspring1->reduce();
    offspring2->reduce();

    /*
    if(debug) 
    {
      ofstream ou1("/tmp/1px.out",ios::app);
      ou1  << " parent1   : " << *parent1 << endl;
      ou1  << " parent2   : " << *parent2 << endl;
      ou1  << " offspring1 : " << *offspring1 << endl;
      ou1  << " offspring2 : " << *offspring2 << endl;
      ou1.close();

    }// end of if
    */

    ChromosomesBag ret(2);
    ret.addChromosomes(offspring1);
    ret.addChromosomes(offspring2);

    if(debug) cout << " 3. I am out of SCPCrossover1PX::apply() " << endl;
    return ret;

}// end of apply()

/*************************************************************************/
/*************************************************************************/

ChromosomesBag SCPCrossover1PX::apply(Chromosomes* par1, Chromosomes * par2, float percent)
{   
    return apply(par1,par2);

}// end of apply()

/*************************************************************************/
/** Returns the index of the first bit from left that two parents differ */
/*************************************************************************/
SCPCrossover1PX::findFirstLeftDifferent( Column *pa1, Column *pa2, int NCols)
{
   int ret=0;
   while( ret<NCols && pa1->test(ret)==pa2->test(ret) ) ret++;
   return (ret>=NCols?NCols-1:ret);

}// end of findFirstLeftDifferent()

/*************************************************************************/
/** Returns the index of the last bit from left that two parents  differ */
/*************************************************************************/
SCPCrossover1PX::findFirstRightDifferent( Column *pa1, Column *pa2, int NCols)
{
   int ret=NCols-1;
   while( ret>-1 && pa1->test(ret)==pa2->test(ret) ) ret--;
   return (ret<0?0:ret);

}// end of findFirstRightDifferent()

/*************************************************************************/
/*************************************************************************/
void SCPCrossover1PX::repair(SetCoverSolution* asol) { asol->restoreRandom(); }

